# future
from __future__ import annotations

# stdlib
from datetime import datetime
import time
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import Type
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import VerifyKey

# relative
from ......util import validate_type
from .....common.message import ImmediateSyftMessageWithReply
from .....common.message import ImmediateSyftMessageWithoutReply
from .....common.uid import UID
from ...exceptions import AuthorizationError
from ...exceptions import InvalidParameterValueError
from ...exceptions import MissingRequestKeyError
from ...exceptions import RequestError
from ...node import DuplicateRequestException
from ...node_table.utils import model_to_json
from ..accept_or_deny_request.accept_or_deny_request_messages import (
    AcceptOrDenyRequestMessage,
)
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from ..node_service import ImmediateNodeServiceWithoutReply
from ..request_answer.request_answer_messages import RequestAnswerMessage
from ..request_answer.request_answer_messages import RequestAnswerResponse
from ..request_handler.request_handler_messages import (
    GetAllRequestHandlersResponseMessage,
)
from ..request_handler.request_handler_messages import GetAllRequestHandlersMessage
from ..request_handler.request_handler_messages import UpdateRequestHandlerMessage
from ..request_receiver.request_receiver_messages import RequestMessage
from ..request_receiver.request_receiver_messages import RequestStatus
from .object_request_messages import CreateBudgetRequestMessage
from .object_request_messages import CreateRequestMessage
from .object_request_messages import CreateRequestResponse
from .object_request_messages import DeleteRequestMessage
from .object_request_messages import DeleteRequestResponse
from .object_request_messages import GetAllRequestsMessage
from .object_request_messages import GetAllRequestsResponseMessage
from .object_request_messages import GetBudgetRequestsMessage
from .object_request_messages import GetBudgetRequestsResponse
from .object_request_messages import GetRequestMessage
from .object_request_messages import GetRequestResponse
from .object_request_messages import GetRequestsMessage
from .object_request_messages import GetRequestsResponse
from .object_request_messages import UpdateRequestMessage
from .object_request_messages import UpdateRequestResponse

if TYPE_CHECKING:
    # relative
    from ....domain import Domain


def create_request_msg(
    msg: CreateRequestMessage,
    node: Domain,
    verify_key: VerifyKey,
) -> CreateRequestResponse:
    # Get Payload Content
    current_user_id = msg.content.get("current_user", None)
    object_id = msg.content.get("object_id", None)
    reason = msg.content.get("reason", None)
    request_type = msg.content.get("request_type", None)
    object_type = msg.content.get("object_type", "storable object")

    users = node.users

    if not current_user_id:
        current_user = users.first(
            verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
        )
    else:
        current_user = users.first(id=current_user_id)

    # since we reject/accept requests based on the ID, we don't want there to be
    # multiple requests with the same ID because this could cause security problems.
    _duplicate_request = node.data_requests.contain(
        object_id=object_id,
        verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8"),
    )

    if _duplicate_request:
        raise DuplicateRequestException(
            f"You have already requested {msg.content['object_id']}"
        )

    # Check if object_id/reason/request_type fields are empty
    missing_paramaters = not object_id or not reason or not request_type
    if missing_paramaters:
        raise MissingRequestKeyError(
            message="Invalid request payload, empty fields (object_id/reason/request_type)!"
        )

    valid_paramaters = request_type == "data" or request_type == "budget"

    if not valid_paramaters:
        raise InvalidParameterValueError(
            message='Request type should be either "data” or “budget”.'
        )

    requests = node.data_requests
    object_uid = UID.from_string(object_id)

    request_obj = requests.create_request(
        user_id=current_user.id,
        user_name=current_user.email,
        object_id=object_id,
        reason=reason,
        request_type=request_type,
        verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8"),
        object_type=object_type,
        tags=node.store.get(object_uid, proxy_only=True)._tags,
    )
    request_json = model_to_json(request_obj)

    return CreateRequestResponse(
        address=msg.reply_to,
        status_code=200,
        content=request_json,
    )


def create_budget_request_msg(
    msg: CreateBudgetRequestMessage,
    node: Domain,
    verify_key: VerifyKey,
) -> None:
    if verify_key is None:
        raise ValueError(
            "Can't process Request service without a given " "verification key"
        )

    # since we reject/accept requests based on the ID, we don't want there to be
    # multiple requests with the same ID because this could cause security problems.
    _duplicate_request = node.data_requests.contain(
        verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8"),
        status="pending",
    )

    if _duplicate_request:
        raise DuplicateRequestException(
            "You have already requested this item before!",
            node.data_requests.all(),
            "My Requests",
        )

    current_user = node.users.first(
        verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
    )

    node.data_requests.create_request(
        user_id=current_user.id,
        user_name=current_user.name,
        user_email=current_user.email,
        user_role=node.roles.first(id=current_user.role).name,
        user_budget=current_user.budget,
        institution=current_user.institution,
        website=current_user.website,
        reason=msg.reason,
        object_id=str(UID().value),
        object_type="<Budget>",
        requested_budget=msg.budget,
        request_type="budget",
        verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8"),
    )


def get_request_msg(
    msg: GetRequestMessage,
    node: Domain,
    verify_key: VerifyKey,
) -> GetRequestResponse:

    # Get Payload Content
    request_id = msg.request_id
    current_user = node.users.first(verify_key=verify_key)
    current_user_id = current_user.id

    users = node.users

    if not current_user_id:
        current_user_id = users.first(
            verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
        ).id

    requests = node.data_requests
    request = requests.first(id=request_id)

    # A user can get a request if he's the owner of that request
    # or has the can_triage_requests permission
    allowed = request.user_id == current_user_id or users.can_triage_requests(
        verify_key=verify_key
    )

    if allowed:
        request_json = model_to_json(request)
    else:
        raise AuthorizationError("You're not allowed to get Request information!")

    return GetRequestResponse(
        address=msg.reply_to,
        status_code=200,
        request_id=request_json,
    )


def get_all_request_msg(
    msg: GetRequestsMessage,
    node: Domain,
    verify_key: VerifyKey,
) -> GetRequestsResponse:
    users = node.users

    allowed = users.can_triage_requests(verify_key=verify_key)

    if allowed:
        requests = node.data_requests.query(request_type="data")
        response = list()
        for request in requests:
            # Get current state user
            if node.data_requests.status(request.id) == RequestStatus.Pending:
                _user = node.users.first(id=request.user_id)
                user = model_to_json(_user)
                user["role"] = node.roles.first(id=_user.role).name
                user["current_budget"] = user["budget"]
            # Get History state user
            else:
                user = node.data_requests.get_user_info(request_id=request.id)
            request = model_to_json(request)
            response.append({"user": user, "req": request})

    else:
        raise AuthorizationError("You're not allowed to get Request information!")

    return GetRequestsResponse(
        status_code=200,
        address=msg.reply_to,
        content=response,
    )


def get_all_budget_requests(
    msg: GetBudgetRequestsMessage,
    node: Domain,
    verify_key: VerifyKey,
) -> GetBudgetRequestsResponse:
    users = node.users

    allowed = users.can_triage_requests(verify_key=verify_key)
    response = list()
    if allowed:
        requests = node.data_requests.query(request_type="budget")
        for request in requests:
            # Get current state user
            if node.data_requests.status(request.id) == RequestStatus.Pending:
                _user = node.users.first(id=request.user_id)
                user = model_to_json(_user)
                user["role"] = node.roles.first(id=_user.role).name
                user["current_budget"] = user["budget"]
                request = model_to_json(request)
            # Get History state user
            else:
                user = node.data_requests.get_user_info(request_id=request.id)
                request = node.data_requests.get_req_info(request_id=request.id)
            response.append({"user": user, "req": request})
    else:
        raise AuthorizationError("You're not allowed to get Request information!")
    return GetBudgetRequestsResponse(
        address=msg.reply_to,
        content=response,
    )


def update_request_msg(
    msg: UpdateRequestMessage,
    node: Domain,
    verify_key: VerifyKey,
) -> UpdateRequestResponse:

    # Get Payload Content
    request_id = msg.request_id
    status = msg.status

    # Check if status field is empty
    missing_paramaters = not status
    if missing_paramaters:
        raise MissingRequestKeyError(
            message="Invalid request payload, empty fields (status)!"
        )

    _req = node.data_requests.first(id=request_id)

    if not _req:
        raise RequestError(message=f"Request ID: {request_id} not found.")

    if status not in ["accepted", "denied"]:
        raise InvalidParameterValueError(
            message='Request status should be either "accepted" or "denied"'
        )

    _can_triage_request = node.users.can_triage_requests(verify_key=verify_key)
    _current_user_key = verify_key.encode(encoder=HexEncoder).decode("utf-8")
    _req_owner = _current_user_key == _req.verify_key

    if status == "accepted" and _can_triage_request:
        # if the type of object being requested has the word 'budget' in it, then we're
        # not really requesting an object per-say, we're requesting for a budget increase
        # TODO: clean up the RequestMessage API to explicitly have multiple request types, including
        # one for budget requests.
        if "<Budget>" in _req.object_type:
            current_user = node.users.first(verify_key=_req.verify_key)
            node.users.set(
                user_id=current_user.id,
                budget=current_user.budget + _req.requested_budget,
            )
        else:
            tmp_obj = node.store.get(UID.from_string(_req.object_id), proxy_only=True)
            tmp_obj.read_permissions[
                VerifyKey(_req.verify_key.encode("utf-8"), encoder=HexEncoder)
            ] = _req.id
            node.store[UID.from_string(_req.object_id)] = tmp_obj
        # this should be an enum not a string
        node.data_requests.set(request_id=_req.id, status=status)  # type: ignore
    elif status == "denied" and (_can_triage_request or _req_owner):
        # this should be an enum not a string
        node.data_requests.set(request_id=_req.id, status=status)  # type: ignore
    else:
        raise AuthorizationError("You're not allowed to update Request information!")

    return UpdateRequestResponse(
        address=msg.reply_to,
        status_code=200,
        status=msg.status,
        request_id=msg.request_id,
    )


def del_request_msg(
    msg: DeleteRequestMessage,
    node: Domain,
    verify_key: VerifyKey,
) -> DeleteRequestResponse:

    request_id = msg.request_id.get("request_id", None)  # type: ignore

    current_user_id = node.users.first(
        verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
    ).id

    requests = node.data_requests
    request = requests.first(id=request_id)

    # Only the creator of a request may delete their request.
    if request.user_id == current_user_id:
        requests.delete(id=request_id)
    else:
        raise AuthorizationError("You're not allowed to delete this Request!")

    return DeleteRequestResponse(
        address=msg.reply_to,
        status_code=200,
        request_id=request_id
        # content={"msg": "Request deleted!"},
    )


def request_answer_msg(
    msg: RequestAnswerMessage, node: Domain, verify_key: VerifyKey
) -> RequestAnswerResponse:
    if verify_key is None:
        raise ValueError(
            "Can't process Request service without a given " "verification key"
        )

    status = node.data_requests.status(request_id=str(msg.request_id.value))
    address = msg.reply_to
    return RequestAnswerResponse(
        request_id=msg.request_id, address=address, status=status
    )


# TODO: Check if this method/message should really be a service_without_reply message
def get_all_requests(
    msg: GetAllRequestsMessage, node: Domain, verify_key: VerifyKey
) -> GetAllRequestsResponseMessage:
    if verify_key is None:
        raise ValueError(
            "Can't process Request service without a given " "verification key"
        )

    _can_triage_request = node.users.can_triage_requests(verify_key=verify_key)

    if _can_triage_request:
        _requests = node.data_requests.all()
    else:
        _requests = node.data_requests.query(
            verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
        )

    data_requests = [
        RequestMessage(
            status=req.status,
            user_name=req.user_name,
            user_email=req.user_email,
            user_role=req.user_role,
            requested_budget=req.requested_budget,
            current_budget=req.user_budget,
            date=str(req.date),
            request_type=req.request_type,
            request_id=UID.from_string(req.id),
            request_description=req.reason,
            address=node.address,
            owner_address=node.address,
            object_id=UID.from_string(req.object_id),
            object_type=req.object_type,
            object_tags=req.tags,
            requester_verify_key=VerifyKey(
                req.verify_key.encode("utf-8"), encoder=HexEncoder
            ),
            timeout_secs=None,
        )
        for req in _requests
    ]

    return GetAllRequestsResponseMessage(requests=data_requests, address=msg.reply_to)


def get_all_request_handlers(
    msg: GetAllRequestHandlersMessage, node: Domain, verify_key: VerifyKey
) -> GetAllRequestHandlersResponseMessage:

    if verify_key is None:
        raise ValueError(
            "Can't process Request service without a given " "verification key"
        )

    handlers = []
    if verify_key == node.root_verify_key:
        existing_handlers = getattr(node, "request_handlers", None)
        if existing_handlers is not None:
            handlers = existing_handlers

    return GetAllRequestHandlersResponseMessage(handlers=handlers, address=msg.reply_to)


class RequestService(ImmediateNodeServiceWithReply):
    INPUT_TYPE = Union[
        Type[CreateRequestMessage],
        Type[GetRequestMessage],
        Type[GetRequestsMessage],
        Type[GetBudgetRequestsMessage],
        Type[UpdateRequestMessage],
        Type[DeleteRequestMessage],
        Type[RequestAnswerMessage],
        Type[GetAllRequestsMessage],
        Type[GetAllRequestHandlersMessage],
    ]

    INPUT_MESSAGES = Union[
        CreateRequestMessage,
        GetRequestMessage,
        GetRequestsMessage,
        GetBudgetRequestsMessage,
        UpdateRequestMessage,
        DeleteRequestMessage,
        RequestAnswerMessage,
        GetAllRequestsMessage,
        GetAllRequestHandlersMessage,
    ]

    OUTPUT_MESSAGES = Union[
        CreateRequestResponse,
        GetRequestResponse,
        GetRequestsResponse,
        GetBudgetRequestsResponse,
        UpdateRequestResponse,
        DeleteRequestResponse,
        RequestAnswerResponse,
        GetAllRequestsResponseMessage,
        GetAllRequestHandlersResponseMessage,
    ]

    msg_handler_map: Dict[INPUT_TYPE, Callable[..., OUTPUT_MESSAGES]] = {
        CreateRequestMessage: create_request_msg,
        GetRequestMessage: get_request_msg,
        GetRequestsMessage: get_all_request_msg,
        GetBudgetRequestsMessage: get_all_budget_requests,
        UpdateRequestMessage: update_request_msg,
        DeleteRequestMessage: del_request_msg,
        RequestAnswerMessage: request_answer_msg,
        GetAllRequestsMessage: get_all_requests,
        GetAllRequestHandlersMessage: get_all_request_handlers,
    }

    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: Domain,
        msg: INPUT_MESSAGES,
        verify_key: VerifyKey,
    ) -> OUTPUT_MESSAGES:
        return RequestService.msg_handler_map[type(msg)](
            msg=msg, node=node, verify_key=verify_key
        )

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
        return [
            CreateRequestMessage,
            GetRequestMessage,
            GetRequestsMessage,
            GetRequestsMessage,
            GetBudgetRequestsMessage,
            UpdateRequestMessage,
            DeleteRequestMessage,
            RequestAnswerMessage,
            GetAllRequestsMessage,
            GetAllRequestHandlersMessage,
        ]


def build_request_message(
    msg: RequestMessage, node: Domain, verify_key: VerifyKey
) -> None:
    if verify_key is None:
        raise ValueError(
            "Can't process Request service without a given " "verification key"
        )
    if msg.requester_verify_key != verify_key:
        raise Exception(
            "You tried to request access for a key that is not yours!"
            "You cannot do this! Whatever key you want to request access"
            "for must be the verify key that also verifies the message"
            "containing the request."
        )

    # since we reject/accept requests based on the ID, we don't want there to be
    # multiple requests with the same ID because this could cause security problems.
    _duplicate_request = node.data_requests.contain(
        object_id=str(msg.object_id.value),
        verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8"),
    )

    if _duplicate_request:
        raise DuplicateRequestException(
            f"You have already requested {msg.object_id} ",
            node.data_requests.all(),
            "My Requests",
        )

    current_user = node.users.first(
        verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
    )

    node.data_requests.create_request(
        user_id=current_user.id,
        user_name=current_user.name,
        user_email=current_user.email,
        user_role=node.roles.first(id=current_user.role).name,
        user_budget=current_user.budget,
        institution=current_user.institution,
        website=current_user.website,
        object_id=str(msg.object_id.value),
        reason=msg.request_description,
        request_type="data",
        verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8"),
        object_type=msg.object_type,
        tags=node.store.get(msg.object_id, proxy_only=True)._tags
        if "budget" not in msg.object_type
        else [],
    )


def accept_or_deny_request(
    msg: AcceptOrDenyRequestMessage, node: Domain, verify_key: VerifyKey
) -> None:
    if verify_key is None:
        raise ValueError(
            "Can't process AcceptOrDenyRequestService without a specified verification key"
        )

    _msg: AcceptOrDenyRequestMessage = validate_type(msg, AcceptOrDenyRequestMessage)

    current_user = node.users.first(
        verify_key=verify_key.encode(encoder=HexEncoder).decode("utf-8")
    )
    # Check if there is any pending request with this id.
    _req = node.data_requests.first(id=str(_msg.request_id.value), status="pending")

    _can_triage_request = node.users.can_triage_requests(verify_key=verify_key)
    if _msg.accept:
        if _req and _can_triage_request:
            if "<Budget>" in _req.object_type:
                current_user = node.users.first(verify_key=_req.verify_key)
                node.users.set(
                    user_id=current_user.id,
                    budget=current_user.budget + _req.requested_budget,
                )
            else:
                tmp_obj = node.store.get(
                    UID.from_string(_req.object_id), proxy_only=True
                )
                tmp_obj.read_permissions[
                    VerifyKey(_req.verify_key.encode("utf-8"), encoder=HexEncoder)
                ] = _req.id
                node.store[UID.from_string(_req.object_id)] = tmp_obj

            # TODO: In the future we'll probably need to keep a request history
            # So, instead of deleting a data access request, we would like to just change its
            # status.
            # node.data_requests.set(request_id=_req.id, status="accepted")
            # node.data_requests.delete(id=_req.id)
            status = "accepted"
    else:
        _req_owner = current_user.verify_key == _req.verify_key
        if _req and (_can_triage_request or _req_owner):
            # TODO: In the future we'll probably need to keep a request history
            # So, instead of deleting a data access request, we would like to just change its
            # status.
            # node.data_requests.set(request_id=_req.id, status="denied")
            # node.data_requests.delete(id=_req.id)
            status = "denied"

    node.data_requests.modify(
        {"id": _req.id},
        {
            "status": status,
            "reviewer_name": current_user.name,
            "reviewer_role": node.roles.first(id=current_user.role).name,
            "reviewer_comment": "",
            "updated_on": datetime.now(),
        },
    )  # type: ignore


def update_req_handler(
    node: Domain,
    msg: UpdateRequestHandlerMessage,
    verify_key: Optional[VerifyKey] = None,
) -> None:
    if verify_key is None:
        raise ValueError(
            "Can't process Request service without a given " "verification key"
        )

    if verify_key == node.root_verify_key:
        replacement_handlers = []

        # find if there exists a handler match the handler passed in
        existing_handlers = getattr(node, "request_handlers", None)
        # debug(f"> Updating Request Handlers with existing: {existing_handlers}")
        if existing_handlers is not None:
            matched = None
            for existing_handler in existing_handlers:
                # we match two handlers according to their tags
                if existing_handler["tags"] == msg.handler["tags"]:
                    matched = existing_handler
                    # if an existing_handler match the passed in handler,
                    # we ignore it in for loop
                    continue
                else:
                    # if an existing_handler does not match the passed in
                    # handler, we keep it
                    replacement_handlers.append(existing_handler)

            if msg.keep:
                msg.handler["created_time"] = time.time()
                replacement_handlers.append(msg.handler)
                if matched is not None:
                    print(
                        f"> Replacing a Request Handler {matched} with: {msg.handler}"
                    )
                    # debug(
                    #    f"> Replacing a Request Handler {matched} with: {msg.handler}"
                    # )
                else:
                    print(f"> Adding a Request Handler {msg.handler}")
                    # debug(f"> Adding a Request Handler {msg.handler}")
            else:
                print(f"> Removing a Request Handler with: {msg.handler}")
                # debug(f"> Removing a Request Handler with: {msg.handler}")

            setattr(node, "request_handlers", replacement_handlers)
            # debug(f"> Finished Updating Request Handlers with: {existing_handlers}")
        else:
            # TODO: Replace line below
            print(f"> Node has no Request Handlers attribute: {type(node)}")
            # error(f"> Node has no Request Handlers attribute: {type(node)}")

    return


class ObjectRequestServiceWithoutReply(ImmediateNodeServiceWithoutReply):
    INPUT_TYPE = Union[
        Type[RequestMessage],
        Type[CreateBudgetRequestMessage],
        Type[AcceptOrDenyRequestMessage],
        Type[UpdateRequestHandlerMessage],
    ]

    INPUT_MESSAGES = Union[
        RequestMessage,
        AcceptOrDenyRequestMessage,
        UpdateRequestHandlerMessage,
    ]

    msg_handler_map: Dict[INPUT_TYPE, Callable[..., None]] = {
        RequestMessage: build_request_message,
        CreateBudgetRequestMessage: create_budget_request_msg,
        AcceptOrDenyRequestMessage: accept_or_deny_request,
        UpdateRequestHandlerMessage: update_req_handler,
    }

    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: Domain,
        msg: Union[
            RequestMessage,
            AcceptOrDenyRequestMessage,
            UpdateRequestHandlerMessage,
        ],
        verify_key: VerifyKey,
    ) -> None:
        ObjectRequestServiceWithoutReply.msg_handler_map[type(msg)](
            msg=msg, node=node, verify_key=verify_key
        )

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithoutReply]]:
        return [
            RequestMessage,
            AcceptOrDenyRequestMessage,
            UpdateRequestHandlerMessage,
            CreateBudgetRequestMessage,
        ]



# stdlib
import time
from typing import List
from typing import Optional

# third party
from nacl.signing import VerifyKey

# relative
from ......logger import traceback_and_raise
from ....abstract.node import AbstractNode
from ...node import DuplicateRequestException
from ..node_service import ImmediateNodeServiceWithoutReply
from .request_receiver_messages import RequestMessage


class RequestReceiverService(ImmediateNodeServiceWithoutReply):
    @staticmethod
    def message_handler_types() -> List[type]:
        return [RequestMessage]

    @staticmethod
    def process(
        node: AbstractNode, msg: RequestMessage, verify_key: Optional[VerifyKey] = None
    ) -> None:
        if verify_key is None:
            traceback_and_raise(
                ValueError(
                    "Can't process Request service without a given " "verification key"
                )
            )

        if msg.requester_verify_key != verify_key:
            traceback_and_raise(
                Exception(
                    "You tried to request access for a key that is not yours!"
                    "You cannot do this! Whatever key you want to request access"
                    "for must be the verify key that also verifies the message"
                    "containing the request."
                )
            )

        # since we reject/accept requests based on the ID, we don't want there to be
        # multiple requests with the same ID because this could cause security problems.
        for req in node.requests:
            # the same user has requested the same object so we raise a
            # DuplicateRequestException
            if (
                req.object_id == msg.object_id
                and req.requester_verify_key == msg.requester_verify_key
            ):
                traceback_and_raise(
                    DuplicateRequestException(
                        f"You have already requested {msg.object_id}"
                    )
                )

        # using the local arrival time we can expire the request
        msg.set_arrival_time(arrival_time=time.time())

        # At receiving a request from DS, we clear it's object_tags, and re-set it as the
        # tags of the requested object. Because the DS may give fake tags.
        while msg.object_tags:
            msg.object_tags.pop()

        if "budget" not in msg.object_type:
            msg.object_tags.extend(node.store.get(msg.object_id, proxy_only=True)._tags)

        node.requests.append(msg)



# stdlib
import sys
from typing import Any
from typing import List
from typing import Optional
from typing import Union

# third party
from pandas import DataFrame

# relative
from .....core.common.uid import UID
from .....grid.client.proxy_client import ProxyClient
from .....lib.python import String
from .....logger import error
from .....logger import start
from .....logger import stop
from ....node.common import AbstractNodeClient
from ..node_service.peer_discovery.peer_discovery_messages import (
    GetPeerInfoMessageWithReply,
)
from ..node_service.peer_discovery.peer_discovery_messages import (
    PeerDiscoveryMessageWithReply,
)
from .request_api import RequestAPI


class DomainRequestAPI(RequestAPI):
    def __init__(self, client: AbstractNodeClient):
        super().__init__(client=client)
        self.cache_time = 0.0
        self.cache: Optional[List[Any]] = None
        self.num_known_domains_even_offline_ones = 0
        self.timeout = 6000  # check for new domains every 100 minutes

    def all(self, pandas: bool = True) -> List[Any]:
        response = self.perform_api_request_generic(
            syft_msg=PeerDiscoveryMessageWithReply, content={}
        )
        result = response.payload.kwargs  # type: ignore

        if result["status"] == "ok":
            data = result["data"]
            # if (
            #     self.cache is None
            #     or (time.time() - self.cache_time > self.timeout)
            #     or len(_data) != self.num_known_domains_even_offline_ones
            # ):
            #     # check for logged in domains if the number of possible domains changes (if a new domain shows up)
            #     self.num_known_domains_even_offline_ones = len(_data)
            #     # n = len(_data)
            #     # data = list()
            #     # args = [
            #     #     (self, i, n, domain_metadata["id"])
            #     #     for i, domain_metadata in enumerate(_data)
            #     # ]

            #     # # Check domain status sequentially
            #     # for i, arg in enumerate(args):
            #     #     if check_domain_status(*arg):
            #     #         data.append(_data[i])

            #     # do not check domain status - assume network will drop stale domains
            #     sys.stdout.write("\r                                             ")

            #     self.cache = _data
            #     self.cache_time = time.time()
            # else:
            #     data = self.cache

            if pandas:
                data = DataFrame(data)

            return data
        return []

    def get(self, key: Union[str, int, UID, String], timeout: Optional[int] = None) -> ProxyClient:  # type: ignore
        # to make sure we target the remote Domain through the proxy we need to
        # construct an 💠 Address which includes the correct UID for the Domain
        # position in the 4 hierarchical locations
        node_uid = key
        try:
            if isinstance(node_uid, int):
                domain_metadata = self.all(pandas=False)[node_uid]
                node_uid = str(domain_metadata["id"])
            elif isinstance(node_uid, String):
                node_uid = node_uid.upcast()
        except Exception as e:
            error(f"Invalid int or String key for list of Domain Clients. {e}")

        if isinstance(node_uid, UID):
            node_uid = node_uid.no_dash

        if not isinstance(node_uid, str):
            msg = (
                f"Unable to get ProxyClient with key with type {type(node_uid)} {node_uid}. "
                "API Request requires key to resolve to a str."
            )
            error(msg)
            raise Exception(msg)
        response = self.perform_api_request_generic(
            syft_msg=GetPeerInfoMessageWithReply,
            content={"uid": node_uid},
            timeout=timeout,
        )

        result = response.payload.kwargs.upcast()  # type: ignore

        # a ProxyClient requires an existing NetworkClient and a Remote Domain Address
        # or a known Domain Node UID, and a Node Name
        proxy_client = ProxyClient.create(
            proxy_node_client=self.client,
            remote_domain=node_uid,
            domain_name=result["data"]["name"],
        )

        return proxy_client

    def __getitem__(self, key: Union[str, int, UID]) -> ProxyClient:
        return self.get(key=key)


def check_domain_status(
    self: DomainRequestAPI, i: int, n: int, domain_uid: str
) -> bool:

    sys.stdout.write(
        "\rChecking whether domains are online: " + str(i + 1) + " of " + str(n)
    )

    status = False
    try:
        stop()
        status = self.get(domain_uid, timeout=1).ping
        start()
    except Exception as e:  # nosec
        # if pinging the domain causes an exception we just wont
        # include it in the array
        print("Error", e)
        pass

    return status



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/connections/http_connection.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n,proto/grid/connections/http_connection.proto\x12\x15syft.grid.connections"Q\n\x12GridHTTPConnection\x12\x10\n\x08\x62\x61se_url\x18\x01 \x01(\t\x12\x15\n\rsession_token\x18\x02 \x01(\t\x12\x12\n\ntoken_type\x18\x03 \x01(\tb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.grid.connections.http_connection_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _GRIDHTTPCONNECTION._serialized_start = 71
    _GRIDHTTPCONNECTION._serialized_end = 152
# @@protoc_insertion_point(module_scope)



# stdlib
import warnings

# its not clear what is causing this warning spam but it must be a third party numerical
# library changing numpy's smallest subnormal or something?
# https://github.com/numpy/numpy/issues/20895
warnings.filterwarnings(
    "ignore",
    ".*The value of the smallest subnormal for.*",
    category=UserWarning,
)



# stdlib
import sys
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

# third party
from nacl.signing import SigningKey
from nacl.signing import VerifyKey
from pandas import DataFrame
from typing_extensions import final

# relative
from ...grid.client.proxy_client import ProxyClient
from ...logger import traceback_and_raise
from ..common.message import SyftMessage
from ..common.uid import UID
from ..io.location import Location
from ..io.location import SpecificLocation
from ..io.route import Route
from .common.action.exception_action import ExceptionMessage
from .common.client import Client
from .common.client_manager.association_api import AssociationRequestAPI
from .common.client_manager.dataset_api import DatasetRequestAPI
from .common.client_manager.domain_api import DomainRequestAPI
from .common.client_manager.role_api import RoleRequestAPI
from .common.client_manager.user_api import UserRequestAPI
from .common.client_manager.vpn_api import VPNAPI
from .common.node_service.generic_payload.messages import GenericPayloadMessageWithReply
from .common.node_service.network_search.network_search_messages import (
    NetworkSearchMessage,
)
from .enums import RequestAPIFields

DEFAULT_SEARCH_TIMEOUT = 120  # seconds


@final
class NetworkClient(Client):

    network: SpecificLocation  # redefine the type of self.vm to not be optional

    def __init__(
        self,
        name: Optional[str],
        routes: List[Route],
        network: SpecificLocation,
        domain: Optional[Location] = None,
        device: Optional[Location] = None,
        vm: Optional[Location] = None,
        signing_key: Optional[SigningKey] = None,
        verify_key: Optional[VerifyKey] = None,
        version: Optional[str] = None,
    ):
        super().__init__(
            name=name,
            routes=routes,
            network=network,
            domain=domain,
            device=device,
            vm=vm,
            signing_key=signing_key,
            verify_key=verify_key,
            version=version,
        )

        self.users = UserRequestAPI(client=self)
        self.roles = RoleRequestAPI(client=self)
        self.association = AssociationRequestAPI(client=self)
        self.datasets = DatasetRequestAPI(client=self)
        self.domains = DomainRequestAPI(client=self)

        self.post_init()

        self.vpn = VPNAPI(client=self)

    @property
    def id(self) -> UID:
        return self.network.id

    @property
    def domain(self) -> Optional[Location]:
        """This client points to a node, if that node lives within a domain
        or is a domain itself, this property will return the Location of that domain
        if it is known by the client."""

        return super().domain

    @domain.setter
    def domain(self, new_domain: Location) -> Optional[Location]:
        """This client points to a node, if that node lives within a domain
        or is a domain itself and we learn the Location of that domain, this setter
        allows us to save the Location of that domain for use later. We use a getter
        (@property) and setter (@set) explicitly because we want all clients
        to efficiently save an address object for use when sending messages to their
        target. That address object will include this information if it is available"""

        traceback_and_raise(
            Exception(
                "This client points to a network, you don't need a Domain Location."
            )
        )

    @property
    def device(self) -> Optional[Location]:
        """This client points to a node, if that node lives within a device
        or is a device itself, this property will return the Location of that device
        if it is known by the client."""

        return super().device

    @device.setter
    def device(self, new_device: Location) -> Optional[Location]:
        """This client points to a node, if that node lives within a device
        or is a device itself and we learn the Location of that device, this setter
        allows us to save the Location of that device for use later. We use a getter
        (@property) and setter (@set) explicitly because we want all clients
        to efficiently save an address object for use when sending messages to their
        target. That address object will include this information if it is available"""

        traceback_and_raise(
            Exception(
                "This client points to a network, you don't need a Device Location."
            )
        )

    @property
    def vm(self) -> Optional[Location]:
        """This client points to a node, if that node lives within a vm
        or is a vm itself, this property will return the Location of that vm
        if it is known by the client."""

        return super().vm

    @vm.setter
    def vm(self, new_vm: Location) -> Optional[Location]:
        """This client points to a node, if that node lives within a vm
        or is a vm itself and we learn the Location of that vm, this setter
        allows us to save the Location of that vm for use later. We use a getter
        (@property) and setter (@set) explicitly because we want all clients
        to efficiently save an address object for use when sending messages to their
        target. That address object will include this information if it is available"""

        traceback_and_raise(
            Exception("This client points to a network, you don't need a VM Location.")
        )

    def __repr__(self) -> str:
        return f"<{type(self).__name__}: {self.name}>"

    def _repr_html_(self) -> str:
        # create repr_html variable first so that the stdout printing is in the right order
        repr_html = self.domains._repr_html_()

        sys.stdout.write("\r\t\t     " + str(self.name) + " network")
        return repr_html

    def __getitem__(self, item: int) -> ProxyClient:
        return self.domains[item]

    def vpn_status(self) -> Dict[str, Any]:
        return self.vpn.get_status()

    def search(
        self,
        query: List,
        pandas: bool = True,
        timeout: Optional[int] = DEFAULT_SEARCH_TIMEOUT,
    ) -> Any:
        response = self._perform_grid_request(
            grid_msg=NetworkSearchMessage, content={"content": query}, timeout=timeout
        )
        result = response.content  # type: ignore
        if result["status"] == "ok":
            result = result["data"]
        else:
            result = []
        if pandas:
            result = DataFrame(result)

        return result

    def _perform_grid_request(
        self,
        grid_msg: Any,
        content: Optional[Dict[Any, Any]] = None,
        timeout: Optional[int] = None,
    ) -> SyftMessage:
        if content is None:
            content = {}
        # Build Syft Message
        content[RequestAPIFields.ADDRESS] = self.address
        content[RequestAPIFields.REPLY_TO] = self.address
        signed_msg = grid_msg(**content).sign(signing_key=self.signing_key)
        # Send to the dest
        response = self.send_immediate_msg_with_reply(msg=signed_msg, timeout=timeout)
        if isinstance(response, ExceptionMessage):
            raise response.exception_type
        else:
            return response

    def perform_api_request_generic(
        self,
        syft_msg: Optional[Type[GenericPayloadMessageWithReply]],
        content: Optional[Dict[Any, Any]] = None,
    ) -> Any:
        if syft_msg is None:
            raise ValueError(
                "Can't perform this type of api request, the message is None."
            )
        else:
            syft_msg_constructor = syft_msg

        if content is None:
            content = {}
        signed_msg = (
            syft_msg_constructor(kwargs=content)
            .to(address=self.address, reply_to=self.address)
            .sign(signing_key=self.signing_key)
        )  # type: ignore
        response = self.send_immediate_msg_with_reply(msg=signed_msg)
        if isinstance(response, ExceptionMessage):
            raise response.exception_type
        else:
            return response



"""Utils functions that might be used into any module."""

# stdlib
from functools import lru_cache
import operator
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import cast

# third party
import numpy as np

# relative
from ..config import DEFAULT_INT_NUMPY_TYPE

RING_SIZE_TO_TYPE: Dict[int, np.dtype] = {
    2**32: np.dtype("int32"),
    2**64: np.dtype("int64"),
    2: np.dtype("bool"),  # Special case: need to do reconstruct and share with XOR
}

TYPE_TO_RING_SIZE: Dict[np.dtype, Optional[int]] = {
    v: k for k, v in RING_SIZE_TO_TYPE.items()
}


def ispointer(obj: Any) -> bool:
    """Check if a given obj is a pointer (is a remote object).
    Args:
        obj (Any): Object.
    Returns:
        bool: True (if pointer) or False (if not).
    """
    if type(obj).__name__.endswith("Pointer") and hasattr(obj, "id_at_location"):
        return True
    return False


@lru_cache()
def get_nr_bits(ring_size: int) -> int:
    """Get number of bits.

    Args:
        ring_size (int): Ring Size.

    Returns:
        int: Bit length.
    """
    return (ring_size - 1).bit_length()


@lru_cache(maxsize=128)
def get_shape(
    op_str: str,
    x_shape: Tuple[int],
    y_shape: Tuple[int],
) -> Tuple[int]:
    """Get the shape of apply an operation on two values

    Args:
        op_str (str): the operation to be applied
        x_shape (Tuple[int]): the shape of op1
        y_shape (Tuple[int]): the shape of op2

    Returns:
        The shape of the result
    """
    if op_str[:2] != "__":
        op = getattr(operator, op_str)
        res = op(np.empty(x_shape), np.empty(y_shape)).shape
    else:
        res = (getattr(np.empty(x_shape), op_str)(np.empty(y_shape))).shape
    res = cast(Tuple[int], res)
    return tuple(res)  # type: ignore


@lru_cache(maxsize=128)
def get_ring_size(
    x_ring_size: int,
    y_ring_size: int,
) -> int:
    """Get the ring_size of apply an operation on two values

    Args:
        x_ring_size (int): the ring size of op1
        y_ring_size (int): the ring size of op2

    Returns:
        The ring size of the result
    """
    if x_ring_size != y_ring_size:
        raise ValueError(
            f"Expected the same ring size for x and y ({x_ring_size} vs {y_ring_size})"
        )

    return x_ring_size


# Code Adapted from Crypten Project:  https://github.com/facebookresearch/CrypTen


def count_wraps(share_list: List[np.ndarray]) -> np.ndarray:
    """Count overflows and underflows if we reconstruct the original value.

    The code is adapted from the Crypten Project.

    Args:
        share_list (List[ShareTensor]): List of the shares.

    Returns:
        torch.Tensor: The number of wraparounds.
    """
    res = np.zeros_like(share_list[0], dtype=DEFAULT_INT_NUMPY_TYPE)
    prev_share = share_list[0]
    for cur_share in share_list[1:]:
        next_share = cur_share + prev_share

        # If prev and current shares are negative,
        # but the result is positive then is an underflow
        res -= (prev_share < 0) & (cur_share < 0) & (next_share > 0)

        # If prev and current shares are positive,
        # but the result is positive then is an overflow
        res += (prev_share > 0) & (cur_share > 0) & (next_share < 0)
        prev_share = next_share

    return res



# DOs and Don's of this class:
# - Do NOT use absolute syft imports (i.e. import syft.core...) Use relative ones.
# - Do NOT put multiple imports on the same line (i.e. from <x> import a, b, c). Use separate lines
# - Do sort imports by length
# - Do group imports by where they come from

# stdlib
from typing import List
from typing import Optional
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from ......logger import debug
from ......logger import traceback_and_raise
from ......util import key_emoji
from ......util import validate_type
from .....common.message import ImmediateSyftMessageWithoutReply
from ....abstract.node import AbstractNode
from ..node_service import ImmediateNodeServiceWithoutReply
from .accept_or_deny_request_messages import AcceptOrDenyRequestMessage


class AcceptOrDenyRequestService(ImmediateNodeServiceWithoutReply):
    @staticmethod
    def process(
        node: AbstractNode,
        msg: ImmediateSyftMessageWithoutReply,
        verify_key: Optional[VerifyKey] = None,
    ) -> None:
        debug((f"> Processing AcceptOrDenyRequestService on {node.pprint}"))

        if verify_key is None:
            traceback_and_raise(
                ValueError(
                    "Can't process AcceptOrDenyRequestService without a "
                    "specified verification key"
                )
            )
        _msg: AcceptOrDenyRequestMessage = validate_type(
            msg, AcceptOrDenyRequestMessage
        )
        if _msg.accept:
            request_id = _msg.request_id
            for req in node.requests:
                if request_id == req.id:
                    # you must be a root user to accept a request
                    if verify_key == node.root_verify_key:
                        print("accepting reqest")
                        obj = node.store.get(req.object_id, proxy_only=True)
                        obj.read_permissions[req.requester_verify_key] = req.id

                        print(obj.read_permissions)

                        # gotta put the object back
                        node.store[req.object_id] = obj

                        print(
                            node.store.get(
                                req.object_id, proxy_only=True
                            ).read_permissions
                        )

                        node.requests.remove(req)

                        debug(f"> Accepting Request:{request_id} {request_id.emoji()}")
                        debug(
                            "> Adding can_read for 🔑 "
                            + f"{key_emoji(key=req.requester_verify_key)} to "
                            + f"Store UID {req.object_id} {req.object_id.emoji()}"
                        )
                        return None

        else:
            request_id = _msg.request_id
            for req in node.requests:
                if request_id == req.id:
                    # if you're a root user you can disable a request
                    # also people can disable their own requets
                    if (
                        verify_key == node.root_verify_key
                        or verify_key == req.requester_verify_key
                    ):
                        node.requests.remove(req)
                        debug(f"> Rejecting Request:{request_id}")
                        return None

    @staticmethod
    def message_handler_types() -> List[Type[AcceptOrDenyRequestMessage]]:
        return [AcceptOrDenyRequestMessage]



# stdlib
from typing import List
from typing import Optional
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from ......util import traceback_and_raise
from ....abstract.node import AbstractNode
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from .simple_messages import SimpleMessage
from .simple_messages import SimpleReplyMessage


class SimpleService(ImmediateNodeServiceWithReply):
    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: AbstractNode,
        msg: SimpleMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> SimpleReplyMessage:
        if verify_key is None:
            traceback_and_raise(
                "Can't process an GetReprService with no verification key."
            )

        result = msg.payload.run(node=node, verify_key=verify_key)

        return SimpleReplyMessage(payload=result, address=msg.reply_to)

    @staticmethod
    def message_handler_types() -> List[Type[SimpleMessage]]:
        return [SimpleMessage]



# stdlib
from typing import List
from typing import Optional
from typing import Type
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import VerifyKey
from pydantic import EmailStr
from typing_extensions import final

# relative
from .....common.serde.serializable import serializable
from ....abstract.node_service_interface import NodeServiceInterface
from ....domain_interface import DomainInterface
from ....domain_msg_registry import DomainMessageRegistry
from ...exceptions import AuthorizationError
from ...exceptions import MissingRequestKeyError
from ...exceptions import UserNotFoundError
from ...node_table.utils import model_to_json
from ...permissions.permissions import BasePermission
from ...permissions.permissions import BinaryOperation
from ...permissions.permissions import UnaryOperation
from ...permissions.user_permissions import IsNodeDaaEnabled
from ...permissions.user_permissions import UserCanCreateUsers
from ...permissions.user_permissions import UserCanTriageRequest
from ...permissions.user_permissions import UserIsOwner
from ..generic_payload.syft_message import NewSyftMessage as SyftMessage
from ..generic_payload.syft_message import ReplyPayload
from ..generic_payload.syft_message import RequestPayload


@serializable(recursive_serde=True)
@final
class CreateUserMessage(SyftMessage, DomainMessageRegistry):

    # Pydantic Inner class to define expected request payload fields.
    class Request(RequestPayload):
        """Payload fields and types used during a User Creation Request."""

        email: EmailStr
        password: str
        name: str
        role: Optional[str] = "Data Scientist"
        institution: Optional[str]
        website: Optional[str]
        budget: Optional[float] = 0.0
        daa_pdf: Optional[bytes] = b""

    # Pydantic Inner class to define expected reply payload fields.
    class Reply(ReplyPayload):
        """Payload fields and types used during a User Creation Response."""

        message: str = "User created successfully!"

    request_payload_type = (
        Request  # Converts generic syft dict into a CreateUserMessage.Request object.
    )
    reply_payload_type = (
        Reply  # Creates a proper Reply payload message structure as a response.
    )

    def run(  # type: ignore
        self, node: DomainInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:  # type: ignore
        """Validates the request parameters and creates a new user.

        Args:
            node (DomainInterface): Domain interface node.
            verify_key (Optional[VerifyKey], optional): User signed verification key. Defaults to None.

        Raises:
            MissingRequestKeyError: If the required request fields are missing.
            AuthorizationError: If user already exists for given email address.

        Returns:
            ReplyPayload: Message on successful user creation.
        """

        # Check if this email was already registered
        try:
            node.users.first(email=self.payload.email)
            # If the email has already been registered, raise exception
            raise AuthorizationError(
                message="You can't create a new User using this email!"
            )
        except UserNotFoundError:
            # If email not registered, a new user can be created.
            pass

        app_id = node.users.create_user_application(
            name=self.payload.name,
            email=self.payload.email,
            password=self.payload.password,
            daa_pdf=self.payload.daa_pdf,
            institution=self.payload.institution,
            website=self.payload.website,
            budget=self.payload.budget,
        )

        node.users.process_user_application(
            candidate_id=app_id, status="accepted", verify_key=verify_key
        )

        return CreateUserMessage.Reply()

    def get_permissions(self) -> List[Type[BasePermission]]:
        """Returns the list of permission classes."""
        return [UserCanCreateUsers, IsNodeDaaEnabled]


@serializable(recursive_serde=True)
@final
class GetUserMessage(SyftMessage, DomainMessageRegistry):

    # Pydantic Inner class to define expected request payload fields.
    class Request(RequestPayload):
        user_id: int

    # Pydantic Inner class to define expected reply payload fields.
    class Reply(ReplyPayload):
        id: int
        name: str
        email: str
        role: Union[int, str]
        budget: float
        created_at: str
        budget_spent: Optional[float] = 0.0
        institution: Optional[str] = ""
        website: Optional[str] = ""
        added_by: Optional[str] = ""

    request_payload_type = Request
    reply_payload_type = Reply

    def run(
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:
        """Returns the user for given user id.

        Args:
            node (NodeServiceInterface): domain/network interface node.
            verify_key (Optional[VerifyKey], optional): user signed verification key. Defaults to None.

        Returns:
            ReplyPayload: Details of the user.
        """
        # Retrieve User Model
        user = node.users.first(id=self.payload.user_id)  # type: ignore

        # Build Reply
        reply = GetUserMessage.Reply(**model_to_json(user))

        # Use role name instead of role ID.
        reply.role = node.roles.first(id=reply.role).name  # type: ignore

        # Get budget spent
        reply.budget_spent = node.users.get_budget_for_user(  # type: ignore
            verify_key=VerifyKey(user.verify_key.encode("utf-8"), encoder=HexEncoder)
        )
        return reply

    def get_permissions(self) -> List[Type[BasePermission]]:
        """Returns the list of permission classes."""
        return [UserCanTriageRequest]


@serializable(recursive_serde=True)
@final
class GetUsersMessage(SyftMessage, DomainMessageRegistry):

    # Pydantic Inner class to define expected request payload fields.
    class Request(RequestPayload):
        pass

    # Pydantic Inner class to define expected reply payload fields.
    class Reply(ReplyPayload):
        users: List[GetUserMessage.Reply] = []

    request_payload_type = Request
    reply_payload_type = Reply

    def run(  # type: ignore
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:
        """Returns the list of all users registered to the node.

        Args:
            node (NodeServiceInterface): domain or network interface node.
            verify_key (Optional[VerifyKey], optional): user signed verification key. Defaults to None.

        Returns:
            ReplyPayload: details of all users registered to the node as a list.
        """
        # Get All Users
        users = node.users.all()
        users_list = list()
        for user in users:
            user_model = GetUserMessage.Reply(**model_to_json(user))

            # Use role name instead of role ID.
            user_model.role = node.roles.first(id=user_model.role).name

            # Remaining Budget
            # TODO:
            # Rename it from budget_spent to remaining budget
            user_model.budget_spent = node.users.get_budget_for_user(  # type: ignore
                verify_key=VerifyKey(
                    user.verify_key.encode("utf-8"), encoder=HexEncoder
                ),
            )
            users_list.append(user_model)

        reply = GetUsersMessage.Reply(users=users_list)
        return reply

    def get_permissions(self) -> List:
        return [UserCanTriageRequest]


@serializable(recursive_serde=True)
@final
class DeleteUserMessage(SyftMessage, DomainMessageRegistry):

    # Pydantic Inner class to define expected request payload fields.
    class Request(RequestPayload):
        user_id: int

    # Pydantic Inner class to define expected reply payload fields.
    class Reply(ReplyPayload):
        message: str = "User deleted successfully!"

    request_payload_type = Request
    reply_payload_type = Reply

    def run(  # type: ignore
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:
        """Delete a user with given user id.

        Args:
            node (NodeServiceInterface): domain or network node interface.
            verify_key (Optional[VerifyKey], optional): user signed verification key. Defaults to None.

        Returns:
            ReplyPayload: message on successful user deletion.
        """

        node.users.delete(id=self.payload.user_id)

        return DeleteUserMessage.Reply()

    def get_permissions(self) -> List[Union[Type[BasePermission], UnaryOperation]]:
        """Returns the list of permission classes applicable to this message."""
        return [UserCanCreateUsers, ~UserIsOwner]


@serializable(recursive_serde=True)
@final
class UpdateUserMessage(SyftMessage, DomainMessageRegistry):

    # Pydantic Inner class to define expected request payload fields.
    class Request(RequestPayload):
        user_id: int
        name: Optional[str] = ""
        email: Optional[EmailStr] = None
        institution: Optional[str] = ""
        website: Optional[str] = ""
        password: Optional[str] = ""
        new_password: Optional[str] = ""
        role: Optional[str] = ""
        budget: Optional[float] = None

    # Pydantic Inner class to define expected reply payload fields.
    class Reply(ReplyPayload):
        message: str = "User updated successfully!"

    request_payload_type = Request
    reply_payload_type = Reply

    def run(  # type: ignore
        self, node: NodeServiceInterface, verify_key: Optional[VerifyKey] = None
    ) -> ReplyPayload:
        """Updates the information for the given user id.

        Args:
            node (NodeServiceInterface): domain or network node interface.
            verify_key (Optional[VerifyKey], optional): user signed verification key. Defaults to None.

        Raises:
            MissingRequestKeyError: If the request parameters are not valid.
            UserNotFoundError: If user with given user id is not present in the database.
            AuthorizationError: If user does not have permission to update their role.

        Returns:
            ReplyPayload: success message on updating the user information.
        """

        _valid_parameters = (
            self.payload.email
            or (self.payload.password and self.payload.new_password)
            or self.payload.role
            or self.payload.name
            or self.payload.institution
            or self.payload.website
        )

        # Change own information
        _valid_user = node.users.contain(id=self.payload.user_id)

        if not _valid_parameters:
            raise MissingRequestKeyError(
                "Missing json fields ( email,password,role,groups, name )"
            )

        if not _valid_user:
            raise UserNotFoundError

        payload_dict = self.payload.dict(exclude_unset=True)
        user_id = payload_dict.pop("user_id")

        # If Change Role Request, then check if user
        # has proper permissions.
        # TODO: This can also be simplified further.
        if self.payload.role:  # type: ignore
            target_user = node.users.first(id=user_id)
            _allowed = (
                self.payload.role != node.roles.owner_role.name  # Target Role != Owner
                and target_user.role
                != node.roles.owner_role.id  # Target User Role != Owner
                and node.users.can_create_users(
                    verify_key=verify_key
                )  # Key Permissions
            )

            # If all premises were respected
            if _allowed:
                role = payload_dict.pop("role")
                new_role_id = node.roles.first(name=role).id
                node.users.set(user_id=user_id, role=new_role_id)  # type: ignore
            elif target_user.role == node.roles.owner_role.id:
                raise AuthorizationError(
                    "You're not allowed to change Owner user roles!"
                )
            else:
                raise AuthorizationError("You're not allowed to change User roles!")

        # Note: Maybe we should create a specific message to change password
        # but in order to accomplish this, we also need to refactory the frontend
        # methods in order to perform a request to the proper endpoint.
        elif self.payload.password and self.payload.new_password:
            node.users.change_password(
                user_id=user_id,
                current_pwd=self.payload.password,
                new_pwd=self.payload.new_password,
            )

            # Delete password keys from the update parameters dictionary to be updated
            # in the next step since we already did it in the previous line.
            del payload_dict["password"]
            del payload_dict["new_password"]

        # Update values of all other parameters
        for param, val in payload_dict.items():
            update_dict = {"user_id": user_id, param: val}
            node.users.set(**update_dict)

        return UpdateUserMessage.Reply()

    def get_permissions(self) -> List[Union[Type[BasePermission], BinaryOperation]]:
        """Returns the list of permission classes applicable to this message."""
        return [UserCanCreateUsers | UserIsOwner]



# stdlib
from typing import Any
from typing import Collection
from typing import Iterable
from typing import List
from typing import Optional
from typing import cast

# third party
from pydantic import BaseSettings
import redis
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import Session

# syft absolute
import syft

# relative
from ....common.uid import UID
from ....node.common.node_table.bin_obj_dataset import BinObjDataset
from ....store import ObjectStore
from ....store.proxy_dataset import ProxyDataset
from ....store.store_interface import StoreKey
from ....store.storeable_object import StorableObject
from ..node_table.bin_obj_metadata import ObjectMetadata


class RedisStore(ObjectStore):
    def __init__(self, db: Session, settings: Optional[BaseSettings] = None) -> None:
        self.db = db
        if settings is None:
            raise Exception("RedisStore requires Settings")
        self.settings = settings
        try:
            self.redis: redis.client.Redis = redis.Redis(
                host=settings.REDIS_HOST,
                port=self.settings.REDIS_PORT,
                db=self.settings.REDIS_STORE_DB_ID,
            )
        except Exception as e:
            print("failed to load redis", e)
            raise e

    def get_or_none(
        self, key: UID, proxy_only: bool = False
    ) -> Optional[StorableObject]:
        try:
            return self.get(key, proxy_only)
        except KeyError:
            return None

    def get_objects_of_type(self, obj_type: type) -> Iterable[StorableObject]:
        # TODO: remove this kind of operation which pulls all the data out in one go
        # raise NotImplementedError("get_objects_of_type")
        # return [obj for obj in self.values() if isinstance(obj.data, obj_type)]
        return self.values()

    def __sizeof__(self) -> int:
        return self.values().__sizeof__()

    def __str__(self) -> str:
        return f"{type(self)}"

    def __len__(self) -> int:
        return self.redis.dbsize()

    def keys(self) -> Collection[UID]:
        key_bytes = self.redis.keys()
        key_ids = [UID.from_string(str(key.decode("utf-8"))) for key in key_bytes]
        return key_ids

    def values(self) -> List[StorableObject]:
        key_bytes = self.redis.keys()
        # this is bad we need to decouple getting the data from the search
        all_values = []
        for key in key_bytes:
            all_values.append(self.get(key))

        return all_values

    def __contains__(self, key: UID) -> bool:
        return str(key.value) in self.redis

    def resolve_proxy_object(self, obj: Any) -> Any:
        obj = obj.get_s3_data(settings=self.settings)
        if obj is None:
            raise Exception(f"Failed to fetch real object from proxy. {type(obj)}")
        return obj

    def get(self, key: StoreKey, proxy_only: bool = False) -> StorableObject:
        key_str, key_uid = self.key_to_str_and_uid(key=key)
        local_session = sessionmaker(bind=self.db)()

        obj = self.redis.get(key_str)
        obj_metadata = (
            local_session.query(ObjectMetadata).filter_by(obj=key_str).first()
        )
        if obj is None or obj_metadata is None:
            raise KeyError(f"Object not found! for UID: {key_str}")

        obj = syft.deserialize(obj, from_bytes=True)
        if proxy_only is False and isinstance(obj, ProxyDataset):
            obj = self.resolve_proxy_object(obj=obj)

        obj = StorableObject(
            id=key_uid,
            data=obj,
            description=obj_metadata.description,
            tags=obj_metadata.tags,
            read_permissions=dict(
                syft.deserialize(
                    bytes.fromhex(obj_metadata.read_permissions), from_bytes=True
                )
            ),
            search_permissions=dict(
                syft.deserialize(
                    bytes.fromhex(obj_metadata.search_permissions), from_bytes=True
                )
            ),
            write_permissions=dict(
                syft.deserialize(
                    bytes.fromhex(obj_metadata.write_permissions), from_bytes=True
                )
            ),
            # name=obj_metadata.name,
        )
        local_session.close()
        return obj

    def is_dataset(self, key: UID) -> bool:
        local_session = sessionmaker(bind=self.db)()
        is_dataset_obj = (
            local_session.query(BinObjDataset).filter_by(obj=str(key.value)).exists()
        )
        is_dataset_obj = local_session.query(is_dataset_obj).scalar()
        local_session.close()
        return is_dataset_obj

    def __getitem__(self, key: StoreKey) -> StorableObject:
        raise Exception("obj = store[key] not allowed because additional args required")

    # allow store[key] = obj
    # but not obj = store[key]
    def __setitem__(self, key: StoreKey, value: StorableObject) -> None:
        self.set(key=key, value=value)

    def set(self, key: StoreKey, value: StorableObject) -> None:
        key_str, _ = self.key_to_str_and_uid(key=key)

        is_proxy_dataset = False
        if isinstance(value._data, ProxyDataset):
            bin = syft.serialize(value._data, to_bytes=True)
            is_proxy_dataset = True
        else:
            bin = syft.serialize(value.data, to_bytes=True)
        self.redis.set(key_str, bin)

        create_metadata = True
        local_session = sessionmaker(bind=self.db)()
        try:
            # use existing metadata row to prevent more than 1
            metadata_obj = (
                local_session.query(ObjectMetadata).filter_by(obj=key_str).all()[-1]
            )
            create_metadata = False
        except Exception:
            pass
            # no metadata row exists lets insert one
            metadata_obj = ObjectMetadata()

        metadata_obj.obj = key_str
        metadata_obj.tags = value.tags
        metadata_obj.description = value.description
        metadata_obj.read_permissions = cast(
            bytes,
            syft.serialize(syft.lib.python.Dict(value.read_permissions), to_bytes=True),
        ).hex()
        metadata_obj.search_permissions = cast(
            bytes,
            syft.serialize(
                syft.lib.python.Dict(value.search_permissions), to_bytes=True
            ),
        ).hex()
        metadata_obj.write_permissions = cast(
            bytes,
            syft.serialize(
                syft.lib.python.Dict(value.write_permissions), to_bytes=True
            ),
        ).hex()
        metadata_obj.is_proxy_dataset = is_proxy_dataset

        local_session = sessionmaker(bind=self.db)()
        if create_metadata:
            local_session.add(metadata_obj)
        local_session.commit()
        local_session.close()

    def delete(self, key: UID) -> None:
        try:
            local_session = sessionmaker(bind=self.db)()
            metadata_to_delete = (
                local_session.query(ObjectMetadata)
                .filter_by(obj=str(key.value))
                .first()
            )
            obj_dataset_to_delete = (
                local_session.query(BinObjDataset).filter_by(obj=str(key.value)).first()
            )
            # Check if the uploaded data is a proxy dataset
            if metadata_to_delete and metadata_to_delete.is_proxy_dataset:
                # Retrieve proxy dataset from store
                obj = self.get(key=key, proxy_only=True)
                proxy_dataset = obj.data
                if proxy_dataset:
                    proxy_dataset.delete_s3_data(settings=self.settings)

            self.redis.delete(str(key.value))
            local_session.delete(metadata_to_delete)
            local_session.delete(obj_dataset_to_delete)
            local_session.commit()
            local_session.close()
        except Exception as e:
            print(f"{type(self)} Exception in __delitem__ error {key}. {e}")

    def clear(self) -> None:
        self.redis.flushdb()
        local_session = sessionmaker(bind=self.db)()
        local_session.query(ObjectMetadata).delete()
        local_session.commit()
        local_session.close()

    def __repr__(self) -> str:
        return f"{type(self)}"
